"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * @since 0.1.0
 */
var Monoid_1 = require("fp-ts/lib/Monoid");
var O = require("fp-ts/lib/Option");
var Semigroup_1 = require("fp-ts/lib/Semigroup");
var Ord_1 = require("fp-ts/lib/Ord");
var pipeable_1 = require("fp-ts/lib/pipeable");
/**
 * 'RetryPolicy' is a 'Monoid'. You can collapse multiple strategies into one using 'concat'.
 * The semantics of this combination are as follows:
 *
 * 1. If either policy returns 'None', the combined policy returns
 * 'None'. This can be used to inhibit after a number of retries,
 * for example.
 *
 * 2. If both policies return a delay, the larger delay will be used.
 * This is quite natural when combining multiple policies to achieve a
 * certain effect.
 *
 * @example
 * import { monoidRetryPolicy, exponentialBackoff, limitRetries } from 'retry-ts'
 *
 * // One can easily define an exponential backoff policy with a limited
 * // number of retries:
 * export const limitedBackoff = monoidRetryPolicy.concat(exponentialBackoff(50), limitRetries(5))
 *
 * @since 0.1.0
 */
exports.monoidRetryPolicy = Monoid_1.getFunctionMonoid(O.getApplyMonoid(__assign({}, Semigroup_1.getJoinSemigroup(Ord_1.ordNumber), { empty: 0 })))();
/**
 * Retry immediately, but only up to `i` times.
 *
 * @since 0.1.0
 */
function limitRetries(i) {
    return function (status) { return (status.iterNumber >= i ? O.none : O.some(0)); };
}
exports.limitRetries = limitRetries;
/**
 * Add an upperbound to a policy such that once the given time-delay
 * amount *per try* has been reached or exceeded, the policy will stop
 * retrying and fail.
 *
 * @since 0.1.0
 */
function limitRetriesByDelay(maxDelay, policy) {
    return function (status) {
        return pipeable_1.pipe(status, policy, O.filter(function (delay) { return delay < maxDelay; }));
    };
}
exports.limitRetriesByDelay = limitRetriesByDelay;
/**
 * Constant delay with unlimited retries
 *
 * @since 0.1.0
 */
function constantDelay(delay) {
    return function () { return O.some(delay); };
}
exports.constantDelay = constantDelay;
/**
 * Set a time-upperbound for any delays that may be directed by the
 * given policy. This function does not terminate the retrying. The policy
 * capDelay(maxDelay, exponentialBackoff(n))` will never stop retrying. It
 * will reach a state where it retries forever with a delay of `maxDelay`
 * between each one. To get termination you need to use one of the
 * 'limitRetries' function variants.
 *
 * @since 0.1.0
 */
function capDelay(maxDelay, policy) {
    return function (status) {
        return pipeable_1.pipe(status, policy, O.map(function (delay) { return Math.min(maxDelay, delay); }));
    };
}
exports.capDelay = capDelay;
/**
 * Grow delay exponentially each iteration.
 * Each delay will increase by a factor of two.
 *
 * @since 0.1.0
 */
function exponentialBackoff(delay) {
    return function (status) { return O.some(delay * Math.pow(2, status.iterNumber)); };
}
exports.exponentialBackoff = exponentialBackoff;
/**
 * Initial, default retry status. Exported mostly to allow user code
 * to test their handlers and retry policies.
 *
 * @since 0.1.0
 */
exports.defaultRetryStatus = {
    iterNumber: 0,
    cumulativeDelay: 0,
    previousDelay: O.none
};
/**
 * Apply policy on status to see what the decision would be.
 *
 * @since 0.1.0
 */
function applyPolicy(policy, status) {
    var previousDelay = policy(status);
    return {
        iterNumber: status.iterNumber + 1,
        cumulativeDelay: status.cumulativeDelay + O.getOrElse(function () { return 0; })(previousDelay),
        previousDelay: previousDelay
    };
}
exports.applyPolicy = applyPolicy;
